"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const assert_1 = __importDefault(require("assert"));
const _1 = require(".");
const client = _1.createClient({
    cluster_id: 0,
    replica_addresses: ['3001']
});
const Zeroed32Bytes = Buffer.alloc(32, 0);
const Zeroed48Bytes = Buffer.alloc(48, 0);
const accountA = {
    id: 17n,
    reserved: Zeroed48Bytes,
    user_data: 0n,
    code: 718,
    unit: 1,
    flags: 0,
    credits_accepted: 0n,
    credits_reserved: 0n,
    debits_accepted: 0n,
    debits_reserved: 0n,
    timestamp: 0n
};
const accountB = {
    id: 19n,
    reserved: Zeroed48Bytes,
    user_data: 0n,
    code: 719,
    unit: 1,
    flags: 0,
    credits_accepted: 0n,
    credits_reserved: 0n,
    debits_accepted: 0n,
    debits_reserved: 0n,
    timestamp: 0n
};
const tests = [];
function test(name, fn) {
    tests.push({ name, fn });
}
test.skip = (name, fn) => {
    console.log(name + ': SKIPPED');
};
test('range checks `unit` and `code` on Account to be u16', async () => {
    const account = { ...accountA, id: 0n, unit: 65535 + 1 };
    const unitError = await client.createAccounts([account]).catch(error => error);
    assert_1.default.strictEqual(unitError.message, 'unit must be a u16.');
    account.unit = 0;
    account.code = 65535 + 1;
    const codeError = await client.createAccounts([account]).catch(error => error);
    assert_1.default.strictEqual(codeError.message, 'code must be a u16.');
    const accounts = await client.lookupAccounts([0n]);
    assert_1.default.strictEqual(accounts.length, 0);
});
test('can create accounts', async () => {
    const errors = await client.createAccounts([accountA]);
    assert_1.default.strictEqual(errors.length, 0);
});
test('can return error', async () => {
    const errors = await client.createAccounts([accountA, accountB]);
    assert_1.default.strictEqual(errors.length, 1);
    assert_1.default.deepStrictEqual(errors[0], { index: 0, code: _1.CreateAccountError.exists });
});
test('throws error if timestamp is not set to 0n', async () => {
    const account = { ...accountA, timestamp: 2n, id: 3n };
    await assert_1.default.rejects(client.createAccounts([account]));
});
test('can lookup accounts', async () => {
    const accounts = await client.lookupAccounts([accountA.id, accountB.id]);
    assert_1.default.strictEqual(accounts.length, 2);
    const account1 = accounts[0];
    assert_1.default.strictEqual(account1.id, 17n);
    assert_1.default.ok(account1.reserved.equals(Zeroed48Bytes));
    assert_1.default.strictEqual(account1.user_data, 0n);
    assert_1.default.strictEqual(account1.code, 718);
    assert_1.default.strictEqual(account1.unit, 1);
    assert_1.default.strictEqual(account1.flags, 0);
    assert_1.default.strictEqual(account1.credits_accepted, 0n);
    assert_1.default.strictEqual(account1.credits_reserved, 0n);
    assert_1.default.strictEqual(account1.debits_accepted, 0n);
    assert_1.default.strictEqual(account1.debits_reserved, 0n);
    assert_1.default.ok(account1.timestamp > 0n);
    const account2 = accounts[1];
    assert_1.default.strictEqual(account2.id, 19n);
    assert_1.default.ok(account2.reserved.equals(Zeroed48Bytes));
    assert_1.default.strictEqual(account2.user_data, 0n);
    assert_1.default.strictEqual(account2.code, 719);
    assert_1.default.strictEqual(account2.unit, 1);
    assert_1.default.strictEqual(account2.flags, 0);
    assert_1.default.strictEqual(account2.credits_accepted, 0n);
    assert_1.default.strictEqual(account2.credits_reserved, 0n);
    assert_1.default.strictEqual(account2.debits_accepted, 0n);
    assert_1.default.strictEqual(account2.debits_reserved, 0n);
    assert_1.default.ok(account2.timestamp > 0n);
});
test('can create a transfer', async () => {
    const transfer = {
        id: 0n,
        amount: 100n,
        code: 1,
        credit_account_id: accountA.id,
        debit_account_id: accountB.id,
        flags: 0,
        user_data: 0n,
        reserved: Zeroed32Bytes,
        timeout: 0n,
        timestamp: 0n,
    };
    const errors = await client.createTransfers([transfer]);
    assert_1.default.strictEqual(errors.length, 0);
    const accounts = await client.lookupAccounts([accountA.id, accountB.id]);
    assert_1.default.strictEqual(accounts.length, 2);
    assert_1.default.strictEqual(accounts[0].credits_accepted, 100n);
    assert_1.default.strictEqual(accounts[0].credits_reserved, 0n);
    assert_1.default.strictEqual(accounts[0].debits_accepted, 0n);
    assert_1.default.strictEqual(accounts[0].debits_reserved, 0n);
    assert_1.default.strictEqual(accounts[1].credits_accepted, 0n);
    assert_1.default.strictEqual(accounts[1].credits_reserved, 0n);
    assert_1.default.strictEqual(accounts[1].debits_accepted, 100n);
    assert_1.default.strictEqual(accounts[1].debits_reserved, 0n);
});
test('can create a two-phase transfer', async () => {
    let flags = 0;
    flags |= _1.TransferFlags.two_phase_commit;
    const transfer = {
        id: 1n,
        amount: 50n,
        code: 1,
        credit_account_id: accountA.id,
        debit_account_id: accountB.id,
        flags,
        user_data: 0n,
        reserved: Zeroed32Bytes,
        timeout: BigInt(2e9),
        timestamp: 0n,
    };
    const errors = await client.createTransfers([transfer]);
    assert_1.default.strictEqual(errors.length, 0);
    const accounts = await client.lookupAccounts([accountA.id, accountB.id]);
    assert_1.default.strictEqual(accounts.length, 2);
    assert_1.default.strictEqual(accounts[0].credits_accepted, 100n);
    assert_1.default.strictEqual(accounts[0].credits_reserved, 50n);
    assert_1.default.strictEqual(accounts[0].debits_accepted, 0n);
    assert_1.default.strictEqual(accounts[0].debits_reserved, 0n);
    assert_1.default.strictEqual(accounts[1].credits_accepted, 0n);
    assert_1.default.strictEqual(accounts[1].credits_reserved, 0n);
    assert_1.default.strictEqual(accounts[1].debits_accepted, 100n);
    assert_1.default.strictEqual(accounts[1].debits_reserved, 50n);
    const transfers = await client.lookupTransfers([transfer.id]);
    assert_1.default.strictEqual(transfers.length, 1);
    assert_1.default.strictEqual(transfers[0].id, 1n);
    assert_1.default.strictEqual(transfers[0].debit_account_id, accountB.id);
    assert_1.default.strictEqual(transfers[0].credit_account_id, accountA.id);
    assert_1.default.strictEqual(transfers[0].user_data, 0n);
    assert_1.default.notStrictEqual(transfers[0].reserved, Zeroed32Bytes);
    assert_1.default.strictEqual(transfers[0].timeout > 0, true);
    assert_1.default.strictEqual(transfers[0].code, 1);
    assert_1.default.strictEqual(transfers[0].flags, 2);
    assert_1.default.strictEqual(transfers[0].amount, 50n);
    assert_1.default.strictEqual(transfers[0].timestamp > 0, true);
});
test('can commit a two-phase transfer', async () => {
    const commit = {
        id: 1n,
        code: 1,
        flags: 0,
        reserved: Zeroed32Bytes,
        timestamp: 0n,
    };
    const errors = await client.commitTransfers([commit]);
    assert_1.default.strictEqual(errors.length, 0);
    const accounts = await client.lookupAccounts([accountA.id, accountB.id]);
    assert_1.default.strictEqual(accounts.length, 2);
    assert_1.default.strictEqual(accounts[0].credits_accepted, 150n);
    assert_1.default.strictEqual(accounts[0].credits_reserved, 0n);
    assert_1.default.strictEqual(accounts[0].debits_accepted, 0n);
    assert_1.default.strictEqual(accounts[0].debits_reserved, 0n);
    assert_1.default.strictEqual(accounts[1].credits_accepted, 0n);
    assert_1.default.strictEqual(accounts[1].credits_reserved, 0n);
    assert_1.default.strictEqual(accounts[1].debits_accepted, 150n);
    assert_1.default.strictEqual(accounts[1].debits_reserved, 0n);
});
test('can reject a two-phase transfer', async () => {
    const transfer = {
        id: 3n,
        amount: 50n,
        code: 1,
        credit_account_id: accountA.id,
        debit_account_id: accountB.id,
        flags: _1.TransferFlags.two_phase_commit,
        user_data: 0n,
        reserved: Zeroed32Bytes,
        timeout: BigInt(1e9),
        timestamp: 0n,
    };
    const transferErrors = await client.createTransfers([transfer]);
    assert_1.default.strictEqual(transferErrors.length, 0);
    const reject = {
        id: 3n,
        code: 1,
        flags: _1.CommitFlags.reject,
        reserved: Zeroed32Bytes,
        timestamp: 0n,
    };
    const errors = await client.commitTransfers([reject]);
    assert_1.default.strictEqual(errors.length, 0);
    const accounts = await client.lookupAccounts([accountA.id, accountB.id]);
    assert_1.default.strictEqual(accounts.length, 2);
    assert_1.default.strictEqual(accounts[0].credits_accepted, 150n);
    assert_1.default.strictEqual(accounts[0].credits_reserved, 0n);
    assert_1.default.strictEqual(accounts[0].debits_accepted, 0n);
    assert_1.default.strictEqual(accounts[0].debits_reserved, 0n);
    assert_1.default.strictEqual(accounts[1].credits_accepted, 0n);
    assert_1.default.strictEqual(accounts[1].credits_reserved, 0n);
    assert_1.default.strictEqual(accounts[1].debits_accepted, 150n);
    assert_1.default.strictEqual(accounts[1].debits_reserved, 0n);
});
test('can link transfers', async () => {
    const transfer1 = {
        id: 4n,
        amount: 100n,
        code: 1,
        credit_account_id: accountA.id,
        debit_account_id: accountB.id,
        flags: _1.TransferFlags.linked,
        user_data: 0n,
        reserved: Zeroed32Bytes,
        timeout: 0n,
        timestamp: 0n,
    };
    const transfer2 = {
        id: 4n,
        amount: 100n,
        code: 1,
        credit_account_id: accountA.id,
        debit_account_id: accountB.id,
        flags: 0,
        user_data: 0n,
        reserved: Zeroed32Bytes,
        timeout: 0n,
        timestamp: 0n,
    };
    const errors = await client.createTransfers([transfer1, transfer2]);
    assert_1.default.strictEqual(errors.length, 2);
    assert_1.default.deepStrictEqual(errors[0], { index: 0, code: _1.CreateTransferError.linked_event_failed });
    assert_1.default.deepStrictEqual(errors[1], { index: 1, code: _1.CreateTransferError.exists_with_different_flags });
    const accounts = await client.lookupAccounts([accountA.id, accountB.id]);
    assert_1.default.strictEqual(accounts.length, 2);
    assert_1.default.strictEqual(accounts[0].credits_accepted, 150n);
    assert_1.default.strictEqual(accounts[0].credits_reserved, 0n);
    assert_1.default.strictEqual(accounts[0].debits_accepted, 0n);
    assert_1.default.strictEqual(accounts[0].debits_reserved, 0n);
    assert_1.default.strictEqual(accounts[1].credits_accepted, 0n);
    assert_1.default.strictEqual(accounts[1].credits_reserved, 0n);
    assert_1.default.strictEqual(accounts[1].debits_accepted, 150n);
    assert_1.default.strictEqual(accounts[1].debits_reserved, 0n);
});
async function main() {
    const start = new Date().getTime();
    try {
        for (let i = 0; i < tests.length; i++) {
            await tests[i].fn().then(() => {
                console.log(tests[i].name + ": PASSED");
            }).catch(error => {
                console.log(tests[i].name + ": FAILED");
                throw error;
            });
        }
        const end = new Date().getTime();
        console.log('Time taken (s):', (end - start) / 1000);
    }
    finally {
        await client.destroy();
    }
}
main().catch((error) => {
    console.log('operator:', error.operator);
    console.log('stack:', error.stack);
});
//# sourceMappingURL=test.js.map